<?php
/**
 * ownCloud
 *
 * @author Owen Winkler
 * @author Vincent Petry <pvince81@owncloud.com>
 * @copyright (C) 2017 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Enterprise_Key;

use OC\HintException;
use OCP\IConfig;
use OCP\IUserSession;
use OCP\IGroupManager;
use Symfony\Component\EventDispatcher\GenericEvent;

use OCP\ILogger;
use OCA\Enterprise_key\Service\AppService;

class EnterpriseKey {
	const WARNING_THRESHOLD = 60;

	/** @var string $org The org name of the key */
	protected $org;
	/** @var int $date A unix timestamp */
	protected $date;
	/** @var  string $rawCodes The raw code string */
	protected $rawCodes;
	/** @var string $codes The code string parsed into CRC chunks */
	protected $codes;
	/** @var string $hash The hash of the keycode */
	protected $hash;
	/** @var IConfig  */
	protected $config;

	/**
	 * @var ILogger
	 */
	protected $logger;

	/**
	 * @var AppService
	 */
	protected $appService;

	/**
	 * Class Constructor
	 * @param bool|string $licVal False to use the config, or a license key string to load
	 * @param IConfig $config
	 */
	public function __construct($licVal = false, IConfig $config, $appService = null) {
		$this->config = $config;
		$licVal = $licVal ?: $this->getLicenseKey();
		list($this->org, $date, $this->rawCodes, $this->hash) = explode('-', $licVal);
		$this->codes = str_split(strtoupper($this->rawCodes), 8);
		$this->date = strtotime(substr($date, 0, 4) . '-' . substr($date, 4, 2) . '-' . substr($date, 6));

		$this->logger = \OC::$server->getLogger();

		if ($appService === null) {
			$appService = new AppService(
				$this->logger,
				\OC::$server->getAppManager(),
				\OC::$server->getAppConfig()
			);
		}
		$this->appService = $appService;
	}

	/**
	 * returns the license key from config.php, returns the license key
	 * from appconfig if it is not available in config.php, returns the
	 * fallback `x-0-x-x` if none is available at all
	 * @return string
	 */
	private function getLicenseKey() {
		$licenseKey = $this->config->getSystemValue('license-key');

		if ($licenseKey) {
			return $licenseKey;
		}

		return $this->config->getAppValue('enterprise_key', 'license-key', 'x-0-x-x');
	}

	public function has($checkVal) {
		$hash = strtoupper(hash("crc32b", $checkVal));
		return in_array($hash, $this->codes);
	}

	/**
	 * Returns whether the key is or was valid, even if expired.
	 *
	 * @return true if the key is or was valid, false otherwise
	 */
	public function isValid() {
		$check = sprintf('%x', crc32(strtolower($this->org) . 'zz' . strtolower($this->rawCodes) . 'zz' . date('Ymd', $this->date)));
		if($check != strtolower($this->hash)
			|| !$this->has(date('Ymd', $this->date))) {
			return false;
		}
		return true;
	}

	/**
	 * Check the license key or look for a specific value in the bloom filter
	 * @param null|string $value Null or a value to search for in the bloom filter
	 * @return bool True if the requested value was found in the bloom filter
	 */
	public function check($value = null) {
		if(!$this->isValid()) {
			if ($this->appService->disableEnterpriseApps()) {
				$l = \OC::$server->getL10N('enterprise_key');
				$err = $l->t('The enterprise license key is invalid, enterprise apps have been disabled.');
				$this->logger->warning($err, ['app' => 'enterprise']);
			}
			return false;
		}
		if($this->date < time() && $this->has('demo') ) {
			if ($this->appService->disableEnterpriseApps()) {
				$l = \OC::$server->getL10N('enterprise_key');
				$err = $l->t('The enterprise license key has expired, enterprise apps have been disabled.');
				$err .= ' ' . $l->t('Please contact sales@owncloud.com for a new license key.');
				$this->logger->warning($err, ['app' => 'enterprise']);
			}
			return false;
		}
		if(!empty($value)) {
			return $this->has($value);
		}
		return true;
	}

	public function expiresInDays() {
		return ceil(($this->date - time()) / 86400);
	}

	/**
	 * Returns whether the key has expired
	 *
	 * @return bool true if the key has expired, false otherwise
	 */
	public function hasExpired() {
		// note: zero days is not expired
		return $this->expiresInDays() < 0;
	}

	/**
	 * @param $event GenericEvent
	 */
	public function getLicenseInfo($event) {
		$licenseKey = $this->config->getSystemValue('license-key', 'x-0-x-x');
		if ( $licenseKey !== 'x-0-x-x') {
			$validUntil = "Valid until " . Date("Y-m-d", strtotime("+" . $this->expiresInDays() . " days"));
			$keyPrefix = explode("-", $this->config->getSystemValue('license-key','x-0-x-x'))[0];
			$event->setArguments(array(['Enterprise License Key' => ['license validity' => $validUntil, 'license-key-prefix' => $keyPrefix]]));
		} else {
			$licKeyUnavailable = "No license key available";
			$event->setArguments(array(['Enterprise License Key' => ['license availability' => $licKeyUnavailable]]));
		}
	}


	/**
	 * @param array $data The array of data supplied by the config.php hook \OCP\Config::js
	 */
	public static function getJs($data) {
		$key = new \OCA\Enterprise_Key\EnterpriseKey(false, \OC::$server->getConfig());
		self::makeLicenseArray(
			$data,
			$key,
			\OC::$server->getUserSession(),
			\OC::$server->getGroupManager()
		);
	}

	/**
	 * Generates messages when appropriate
	 *
	 * @param array $data JS array
	 * @param \OCA\Enterprise_Key\EnterpriseKey $key enterprise key
	 * @param IUserSession $userSession
	 * @param IGroupManager $groupManager
	 * @return string|null notice text or null if nothing needs to be shown
	 */
	public static function makeLicenseArray(&$data, EnterpriseKey $key, IUserSession $userSession, IGroupManager $groupManager) {
		// FIXME: this really should be in its own class
		$user = null;
		if ($userSession !== null) {
			$user = $userSession->getUser();
		}

		// seems unused ?
		$data['array']['licenseKeyDaysLeft'] = $key->expiresInDays();

		// regular logged in user sees no license message
		if (!is_null($user) && !$groupManager->isAdmin($user->getUID())) {
			return;
		}

		// note: this message will be used for both the login page and banner notification
		$message = self::getLicenseKeyNotice($key, !is_null($user));
		if (!is_null($message)) {
			$data['array']['licenseKeyNotice'] = '"' . addslashes($message) . '"';
		}
	}

	/**
	 * Returns the notice
	 *
	 * @param \OCA\Enterprise_Key\EnterpriseKey $key enterprise key
	 * @param bool $isLoggedInAsAdmin whether the user is logged in as admin or logged out
	 * @return string|null notice text or null if nothing needs to be shown
	 */
	public static function getLicenseKeyNotice($key, $isLoggedInAsAdmin) {
		$l = \OC::$server->getL10N('enterprise_key');
		$contactMessage = $l->t('Please contact your administrator or <a href="mailto:sales@owncloud.com?subject=Renew+ownCloud+License">sales@owncloud.com</a> for a new license key.');
		if (!$key->isValid()) {
			return $l->t('Invalid license key !') . '<br/>' . $contactMessage;
		}
		$isDemo = $key->has('demo');
		if ($key->hasExpired()) {
			$expireMessage = $l->t('Your Enterprise license key has expired.') . '<br/>';
			$appDisabledMessage = '';
			if ($isDemo) {
				$appDisabledMessage = $l->t('Enterprise features have been disabled.') . '<br/>';
			}
			return $expireMessage . $appDisabledMessage . $contactMessage;
		}

		$expiresIn = $key->expiresInDays();

		// for demo keys, only show evaluation message on the login page
		if (!$isLoggedInAsAdmin && $isDemo) {
			return '<b>' . $l->t('Evaluation - expires in %d days.', array($expiresIn)) . '</b><br>';
		}

		// for regular keys, show a notification banner when logged in, when expiration is close
		if ($isLoggedInAsAdmin && !$isDemo && $expiresIn <= self::WARNING_THRESHOLD) {
			return
				$l->t('Your Enterprise license key is about to expire (days remaining: %d).', array($expiresIn)) . '<br/>' .
				$contactMessage;
		}

		return null;
	}
}
