<?php
/**
 * @author Jesus Macias <jesus@owncloud.com>
 *
 * @copyright Copyright (c) 2016, ownCloud, Inc.
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License, version 3,
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 *
 */

namespace OCA\windows_network_drive\lib;

use Icewind\SMB\Exception\ForbiddenException;
use OCA\Files_External\Lib\Storage\SMB;
use OCP\Files\StorageNotAvailableException;

/**
 * FS class to connect to windows network drives
 */
class WND extends SMB {
	protected $isInitialized = false;
	protected $shareName;
	protected $host;
	protected $user;
	protected $password;

	protected $wndNotifier;

	public function __construct($params) {
		//Add domain/workgroup to user param
		if (isset($params['domain'])) {
			$params['user'] = Utils::conditionalDomainPlusUsername($params['domain'], $params['user']);
		}
		parent::__construct($params);

		$this->user = $params['user'];
		$this->password = $params['password'];
		$this->shareName = $params['share'];
		$this->host = $params['host'];

		// register in the singleton notifier
		$this->wndNotifier = WNDNotifier::getSingleton();
		$this->wndNotifier->registerWND($this);
	}

	public function init() {
		if ($this->isInitialized) {
			return;
		}

		if ($this->password === '' && $this->user !== '') {
			throw new StorageNotAvailableException('Password required when username given');
		}

		try {
			$this->connectionTestWithReset();
		} catch (ForbiddenException $e) {
			throw new StorageNotAvailableException('Storage not available, perhaps outdated credentials.');
		}
		$this->isInitialized = true;
	}

	/**
	 * Get the username trying to access to the windows drive. It might be different than the
	 * current ownCloud user
	 *
	 * @return string the username
	 */
	public function getUser() {
		$this->log('getUser: '.$this->user, \OCP\Util::DEBUG);
		return $this->user;
	}

	/**
	 * Get the host we're trying to access
	 *
	 * @return string the host
	 */
	public function getHost() {
		$this->log('getHost: '.$this->host, \OCP\Util::DEBUG);
		return $this->host;
	}

	/**
	 * Get the sharename where we're trying to access
	 *
	 * @return string the share name
	 */
	public function getShareName() {
		$this->log('getShareName: '.$this->shareName, \OCP\Util::DEBUG);
		return $this->shareName;
	}

	/**
	 * Get the folder that will act as our root
	 *
	 * @return string the root folder
	 */
	public function getRoot() {
		$this->log('getRoot: '.$this->root, \OCP\Util::DEBUG);
		return $this->root;
	}

	public function getPassword() {
		$this->log('getPassword: (fetched)', \OCP\Util::DEBUG);
		return $this->password;
	}

	/**
	 * Get the FS id (for ownCloud purposes). Method overloaded to keep retrocompatibility.
	 *
	 * @return string the FS id
	 */
	public function getId() {
		$id = 'wnd::' . $this->getUser() . '@' . $this->getHost() . '/' . $this->getShareName() . '/' . $this->getRoot();
		$this->log('getId id: '.$id, \OCP\Util::DEBUG);
		return $id;
	}

	/**
	 * @param string $message
	 */
	private function log($message, $level, $from='wnd') {
		if (\OC::$server->getConfig()->getSystemValue('wnd.logging.enable', false) === true) {
			\OCP\Util::writeLog($from, $message, $level);
		}
	}

	protected function getFileInfo($path) {
		$this->init();
		return parent::getFileInfo($path);
	}

	protected function getFolderContents($path) {
		$this->init();
		return parent::getFolderContents($path);
	}

	public function unlink($path) {
		$this->init();
		return parent::unlink($path);
	}

	public function fopen($path, $mode) {
		$this->init();
		return parent::fopen($path, $mode);
	}

	public function rmdir($path) {
		$this->init();
		return parent::rmdir($path);
	}

	public function mkdir($path) {
		$this->init();
		return parent::mkdir($path);
	}

	public function touch($path, $time=null) {
		$this->init();
		return parent::touch($path, $time);
	}

	/**
	 * Need direct access to the share because the default function will provoke infinite recursion
	 * in this particular case
	 */
	public function test($personal = false, $testOnly = false) {
		if ($this->password === '' && $this->user !== '') {
			throw new StorageNotAvailableException('Password required when username given');
		}
		if ($testOnly) {
			return $this->connectionTestWithoutReset();
		}
		return $this->connectionTestWithReset();
	}

	/**
	 * Test the connection without resetting the password
	 */
	private function connectionTestWithoutReset() {
		$info = $this->share->stat($this->buildPath(''));
		if ($info->isHidden()) {
			throw new \Icewind\SMB\Exception\ConnectException('Cannot connect to storage because the root folder is hidden');
		}
		return true;
	}

	/**
	 * Test the connection and reset the password if needed
	 */
	private function connectionTestWithReset() {
		try {
			$info = $this->share->stat($this->buildPath(''));
			if ($info->isHidden()) {
				throw new \Icewind\SMB\Exception\ConnectException('Cannot connect to storage because the root folder is hidden');
			}
			return true;
		} catch (ForbiddenException $e) {
			if ($this->password !== '') {
				// FIXME: We need credential information to skip the password reset if login credentials
				// are used or make sure login credentials are reset too
				Utils::resetPassword($this);
				// notify all WND objects to reset the password if needed
				// this object will also be notified
				$this->wndNotifier->notifyChange($this, WNDNotifier::NOTIFY_PASSWORD_REMOVAL);
			}
			throw $e;
		}
	}

	/**
	 * receive notification from a WND object. This might remove the current password if host and
	 * user are the same
	 */
	public function receiveNotificationFrom(WND $other, $changeType) {
		$this->log('notification from ' . $other->getHost() . ' ' . $other->getUser() .' to ' . $this->getHost() . ' ' . $this->getUser(), \OCP\Util::WARN);
		if ($changeType === WNDNotifier::NOTIFY_PASSWORD_REMOVAL &&
				$this->getHost() === $other->getHost() &&
				$this->getUser() === $other->getUser()) {
			$this->password = '';
			$this->isInitialized = false;
		}
	}
}
